unit TGThematic;

interface

Uses WinTypes, WinProcs, SysUtils, Classes, Graphics, Globe3;

type
	EThematicException = class(Exception);

	TThematicBreakType = (btRange, btStdDev, btPercentile);

	TThematicBreak = class(TObject)
	public
		Limit : Double;
		Size : integer;
		Color : TColor;
		Selected : Boolean;

		constructor Create;
	end;

	TGlobeThematic = class( TObject )
	private
		FeNoDataValue : Extended;
		FeStdDev, FeMean, FeSum, FeMin, FeMax : Extended;

		FBreakList : TList;

		FNoDataColor: TColor;
		FStartColor: TColor;
		FEndColor: TColor;
		FPresenter : TGlobePresenter;
	protected
		function GetBreak( iIndex : integer ) : TThematicBreak; virtual;
		function GetBreakCount : integer; virtual;
		procedure SetBreakCount( Value : integer ); virtual;
	public
		constructor Create( ParentPresenter : TGlobePresenter );
		destructor Destroy; override;

		procedure GradientColors; virtual;
		procedure CalcStatistics; virtual;
		procedure CalcBreakLimits( BreakType : TThematicBreakType ); virtual;
		function ValueToBreak(Value: Extended): TThematicBreak;
		procedure SelectBreak( Index : integer );

		property MeanValue: Extended read FeMean;
		property StdDevValue: Extended read FeStdDev;
		property MinValue: Extended read FeMin;
		property MaxValue: Extended read FeMax;
		property SumValue: Extended read FeSum;

		property Breaks[index: Integer]: TThematicBreak read GetBreak;
		property BreakList : TList read FBreakList;
		property BreakCount : integer read GetBreakCount write SetBreakCount;
		property Presenter : TGLobePresenter read FPresenter;
	published
		property NoDataValue: Extended read FeNoDataValue write FeNoDataValue;
		property NoDataColor: TColor read FNoDataColor write FNoDataColor;
		property StartColor: TColor read FStartColor write FStartColor;
		property EndColor: TColor read FEndColor write FEndColor;
	end;

implementation

{------------------------------------------------------------------------------
	TThematicBreak.Create
------------------------------------------------------------------------------}
constructor TThematicBreak.Create;
begin
	inherited Create;

	Limit := 0;
	Size := 0;
	Color := clRed;
	Selected := False;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.Create
------------------------------------------------------------------------------}
constructor TGlobeThematic.Create( ParentPresenter : TGlobePresenter );
begin
	inherited Create;

	FBreakList := TList.Create;
	StartColor := clBlue;
	EndColor := clRed;
	NoDataColor := clBlack;
	FPresenter := ParentPresenter;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.Destroy
------------------------------------------------------------------------------}
destructor TGlobeThematic.Destroy;
begin
	FBreakList.Free;
	inherited Destroy;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.GradientColors
------------------------------------------------------------------------------}
procedure TGlobeThematic.GradientColors;
var
	R, G, B : Byte;
	idx, iSteps : integer;
begin
	iSteps := BreakList.Count - 1;

	for idx := 0 to iSteps do
	begin
		R := GetRValue(StartColor);
		R := R + MulDiv( idx, GetRValue(EndColor) - R, iSteps );
		G := GetGValue(StartColor);
		G := G + MulDiv( idx, GetGValue(EndColor) - G, iSteps );
		B := GetBValue(StartColor);
		B := B + MulDiv( idx, GetBValue(EndColor) - B, iSteps );

		Breaks[idx].Color := RGB(R, G, B);
	end;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.GetBreak
------------------------------------------------------------------------------}
function TGlobeThematic.GetBreak( iIndex : integer ) : TThematicBreak;
begin
	if ( iIndex < 0 ) or ( iIndex >= FBreakList.Count ) then
		raise EThematicException.Create( 'Break Index out of bounds' );

	Result := TThematicBreak( FBreakList[iIndex] );
end;

{------------------------------------------------------------------------------
	TGlobeThematic.GetBreakCount
------------------------------------------------------------------------------}
function TGlobeThematic.GetBreakCount : integer;
begin
	Result := BreakList.Count;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.SetBreakCount
------------------------------------------------------------------------------}
procedure TGlobeThematic.SetBreakCount( Value : integer );
var
	TmpBreak : TThematicBreak;
begin
	with BreakList do
		if Value <> Count then
		begin
			while Count < Value do
				BreakList.Add( TThematicBreak.Create );
			while Count > Value do
			begin
				TmpBreak := BreakList[Count - 1];
				BreakList.Remove( TmpBreak );
				TmpBreak.Free;
			end;
		end;
	SelectBreak( -1 );	{ Clear the Selected flag from the breaks }
end;

{------------------------------------------------------------------------------
	SortByValue
------------------------------------------------------------------------------}
function SortByValue(Item1, Item2: Pointer): Integer;
begin
	if TGlobeObject( Item1 ).Value < TGlobeObject( Item2 ).Value then
		Result := -1
	else
		if TGlobeObject( Item1 ).Value > TGlobeObject( Item2 ).Value then
			Result := 1
		else
			Result := 0;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.CalcBreakLimits
------------------------------------------------------------------------------}
procedure TGlobeThematic.CalcBreakLimits( BreakType : TThematicBreakType );
var
	eStep, eStart : Extended;
	idx, iBreakCount, iFirst : integer;
	SortedList : TList;
begin
	iBreakCount := BreakList.Count - 1;
	case BreakType of
	btRange:
		begin
			eStep := (FeMax - FeMin) / iBreakCount;
			for idx := 0 to iBreakCount do
			begin
				Breaks[idx].Limit := FeMin + eStep * idx;
//				Breaks[idx].Size := StartSize + MulDiv( EndSize - StartSize, idx, iBreakCount );
			end;
		end;
	btStdDev:
		begin
			eStep := FeStdDev;
			eStart := FeMean - (iBreakCount div 2) * FeStdDev;
			for idx := 0 to iBreakCount do
			begin
				Breaks[idx].Limit := eStart + eStep * idx;
//				Breaks[idx].Size := StartSize + MulDiv( EndSize - StartSize, idx, iBreakCount );
			end;
		end;
	btPercentile:
		begin
			SortedList := TList.Create;

			for idx := 0 to Presenter.Layer.ObjectCount - 1 do
				if Presenter.Layer[idx].Presenter = Presenter then
					if Presenter.Layer[idx].Value <> NoDataValue then
						SortedList.Add( Presenter.Layer[idx] );

			SortedList.Sort( SortByValue );

			if SortedList.Count > 0 then
			begin
				for idx := 0 to iBreakCount - 1 do
				begin
					Breaks[idx].Limit := TGlobeObject( SortedList[(SortedList.Count div iBreakCount) * idx]).Value;
//					Breaks[idx].Size := StartSize + MulDiv( EndSize - StartSize, idx, iBreakCount );
				end;
				Breaks[iBreakCount].Limit := TGlobeObject( SortedList[SortedList.Count - 1]).Value;
//				Breaks[iBreakCount].Size := EndSize;
			end;
			SortedList.Free;
		end;
	end;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.ValueToBreak
------------------------------------------------------------------------------}
function TGlobeThematic.ValueToBreak(Value: Extended): TThematicBreak;
var
	idx: Integer;
begin
	Result := nil;
	for idx := 0 to BreakCount - 1 do
		if Value <= Breaks[idx].Limit then
		begin
			Result := Breaks[idx];
			Exit;
		end;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.SelectBreak
------------------------------------------------------------------------------}
procedure TGlobeThematic.SelectBreak( Index : integer );
var
	idx: Integer;
begin
	for idx := 0 to BreakCount - 1 do
		Breaks[idx].Selected := idx = Index;
end;

{------------------------------------------------------------------------------
	TGlobeThematic.CalcStatistics
------------------------------------------------------------------------------}
procedure TGlobeThematic.CalcStatistics;
var
	eTmp : Extended;
	idx, iDataCount : integer;
begin
	{ scan the objects to find the range of values }
	FeMax := -1.1E104932;
	FeMin := 1.1E104932;

	FeSum := 0;
	iDataCount := 0;
	for idx := 0 to Presenter.Layer.ObjectCount - 1 do
		if Presenter.Layer[idx].Presenter = Presenter then
		begin
			eTmp := Presenter.Layer[idx].Value;

			{ update the range values }
			if eTmp <> FeNoDataValue then
			begin
				if eTmp < FeMin then FeMin := eTmp;
				if eTmp > FeMax then FeMax := eTmp;
				FeSum := FeSum + eTmp;
				Inc(iDataCount);
			end;
		end;

	if iDataCount > 0 then
	begin
		FeMean := FeSum / iDataCount;

		{ Calculate the Standard Deviation }
		eTmp := 0;
		for idx := 0 to Presenter.Layer.ObjectCount - 1 do
			if Presenter.Layer[idx].Presenter = Presenter then
				if Presenter.Layer[idx].Value <> FeNoDataValue then
					eTmp := eTmp + Sqr( Presenter.Layer[idx].Value - FeMean );

		FeStdDev := Sqrt( eTmp / iDataCount);
	end;
end;

end.
